import logging
from typing import Callable, Dict
import asyncio
from pydantic import BaseModel, Field

from fluss.api.schema import (
    ArgNodeFragment,
    ArkitektNodeFragment,
    FlowFragment,
    LocalNodeFragment,
    ReactiveNodeFragment,
    ReturnNodeFragment,
    arun,
    asnapshot,
    atrack,
)
from reaktion.atoms.transport import AtomTransport

from reaktion.atoms.utils import atomify
from reaktion.contractors import NodeContractor, arkicontractor, localcontractor
from reaktion.events import EventType, InEvent, OutEvent

from reaktion.utils import connected_events
from rekuest.actors.base import Actor
from rekuest.api.schema import (
    AssignationStatus,
    ProvisionStatus,
    ReservationFragment,
    ReservationStatus,
)
from rekuest.messages import Assignation, Provision
from rekuest.postmans.utils import RPCContract
from typing import Any

logger = logging.getLogger(__name__)

print(asyncio)


class NodeState(BaseModel):
    latestevent: OutEvent


class FlowActor(Actor):
    is_generator: bool = False
    flow: FlowFragment
    agent: Any
    contracts: Dict[str, RPCContract] = Field(default_factory=dict)
    local_contracts: Dict[str, RPCContract] = Field(default_factory=dict)
    expand_inputs: bool = False
    shrink_outputs: bool = False
    provided = False
    is_generator: bool = False
    arkitekt_contractor: NodeContractor = arkicontractor
    local_contractor: NodeContractor = localcontractor

    # Functionality for running the flow

    # Assign Related Functionality
    run_mutation: Callable = arun
    snapshot_mutation: Callable = asnapshot
    track_mutation: Callable = atrack

    atomifier: Callable = atomify
    """ Atomifier is a function that takes a node and returns an atom """

    run_states: Dict[
        str,
        Dict[str, NodeState],
    ] = Field(default_factory=dict)

    reservation_state: Dict[str, ReservationFragment] = Field(default_factory=dict)
    _lock = None

    async def on_provide(self, provision: Provision):
        self._lock = asyncio.Lock()

        [x for x in self.flow.graph.nodes if isinstance(x, ArgNodeFragment)][0]
        [x for x in self.flow.graph.nodes if isinstance(x, ReturnNodeFragment)][0]

        arkitektNodes = [
            x for x in self.flow.graph.nodes if isinstance(x, ArkitektNodeFragment)
        ]

        localNodes = [
            x for x in self.flow.graph.nodes if isinstance(x, LocalNodeFragment)
        ]

        self.contracts = {
            node.id: await self.arkitekt_contractor(node, self)
            for node in arkitektNodes
        }

        self.local_contracts = {
            node.id: await self.local_contractor(node, self) for node in localNodes
        }

        print("Entering Contracts")
        futures = [contract.aenter() for contract in self.contracts.values()]
        await asyncio.gather(*futures)

        print("Entering Local Contracts")
        futures = [contract.aenter() for contract in self.local_contracts.values()]
        await asyncio.gather(*futures)

        self.provided = True
        await self.on_reservation_change(None)

    async def on_reservation_change(self, status: ReservationStatus):
        async with self._lock:
            unactive_reservations = [
                res
                for res in self.contracts.values()
                if res.state != ReservationStatus.ACTIVE
            ]
            if self.provided:
                if len(unactive_reservations) > 0:
                    await self.transport.change_provision(
                        self.provision.provision,
                        status=ProvisionStatus.CRITICAL,
                    )
                else:
                    await self.transport.change_provision(
                        self.provision.provision,
                        status=ProvisionStatus.ACTIVE,
                    )

    async def on_assign(self, assignation: Assignation):
        run = await self.run_mutation(
            assignation=assignation.assignation, flow=self.flow
        )

        await self.aass_log(assignation.assignation, "Starting")

        t = 0
        state = {}
        await self.snapshot_mutation(run=run, events=list(state.values()), t=t)

        try:
            event_queue = asyncio.Queue()

            atomtransport = AtomTransport(queue=event_queue)

            argNode = [
                x for x in self.flow.graph.nodes if isinstance(x, ArgNodeFragment)
            ][0]
            returnNode = [
                x for x in self.flow.graph.nodes if isinstance(x, ReturnNodeFragment)
            ][0]

            participatingNodes = [
                x
                for x in self.flow.graph.nodes
                if isinstance(x, ArkitektNodeFragment)
                or isinstance(x, ReactiveNodeFragment)
                or isinstance(x, LocalNodeFragment)
            ]

            await self.aass_log(assignation.assignation, "Set up the graph")

            async def ass_log(assignation: Assignation, level, message):
                await self.aass_log(assignation.assignation, level, message)
                logging.info(f"{assignation}, {message}")

            atoms = {
                x.id: self.atomifier(
                    x,
                    atomtransport,
                    self.contracts,
                    self.local_contracts,
                    assignation,
                    alog=ass_log,
                )
                for x in participatingNodes
            }

            await self.aass_log(assignation.assignation, "Atomification complete")

            print("Enterying ")
            await asyncio.gather(*[atom.aenter() for atom in atoms.values()])
            print("Enterying complete")

            tasks = [asyncio.create_task(atom.start()) for atom in atoms.values()]

            stream = argNode.outstream[0]
            value = [assignation.args[key] for key, index in enumerate(stream)]

            initial_event = OutEvent(
                handle="return_0",
                type=EventType.NEXT,
                source=argNode.id,
                value=value,
                caused_by=[t],
            )
            initial_done_event = OutEvent(
                handle="return_0",
                type=EventType.COMPLETE,
                source=argNode.id,
                caused_by=[t],
            )

            logger.info(f"Putting initial event {initial_event}")

            await event_queue.put(initial_event)
            await event_queue.put(initial_done_event)

            edge_targets = [e.target for e in self.flow.graph.edges]
            nodes_without_instream = [
                x
                for x in participatingNodes
                if len(x.instream[0]) == 0 and x.id not in edge_targets
            ]

            logger.error(f"Nodes without instream: {nodes_without_instream}")
            for node in nodes_without_instream:
                assert node.id in atoms, "Atom not found. Should not happen."
                atom = atoms[node.id]

                initial_event = InEvent(
                    target=node.id,
                    handle="arg_0",
                    type=EventType.NEXT,
                    value=[],
                    current_t=t,
                )
                done_event = InEvent(
                    target=node.id,
                    handle="arg_0",
                    type=EventType.COMPLETE,
                    current_t=t,
                )

                await atom.put(initial_event)
                await atom.put(done_event)

            complete = False

            returns = []

            while not complete:
                event: OutEvent = await event_queue.get()
                event_queue.task_done()

                if self.flow.brittle:
                    if event.type == EventType.ERROR:
                        raise event.value

                track = await self.track_mutation(
                    run=run,
                    source=event.source,
                    handle=event.handle,
                    caused_by=event.caused_by,
                    value=event.value
                    if event.value and not isinstance(event.value, Exception)
                    else str(event.value),
                    type=event.type,
                    t=t,
                )
                state[event.source] = track.id

                # We tracked the events and proceed

                if t % 3 == 0:
                    await self.snapshot_mutation(
                        run=run, events=list(state.values()), t=t
                    )

                # Creat new events with the new timepoint
                spawned_events = connected_events(self.flow.graph, event, t)
                # Increment timepoint
                t += 1
                # needs to be the old one for now
                if not spawned_events:
                    logger.warning(f"No events spawned from {event}")

                for spawned_event in spawned_events:
                    logger.info(f"-> {spawned_event}")

                    if spawned_event.target == returnNode.id:
                        if spawned_event.type == EventType.NEXT:
                            returns = spawned_event.value
                            if self.is_generator:
                                await self.transport.change_assignation(
                                    assignation.assignation,
                                    status=AssignationStatus.YIELD,
                                    returns=returns,
                                )

                        if spawned_event.type == EventType.ERROR:
                            raise spawned_event.value

                        if spawned_event.type == EventType.COMPLETE:
                            complete = True
                            if not self.is_generator:
                                await self.transport.change_assignation(
                                    assignation.assignation,
                                    status=AssignationStatus.RETURNED,
                                    returns=returns,
                                )
                            else:
                                await self.transport.change_assignation(
                                    assignation.assignation,
                                    status=AssignationStatus.DONE,
                                )

                    else:
                        assert (
                            spawned_event.target in atoms
                        ), "Unknown target. Your flow is connected wrong"
                        if spawned_event.target in atoms:
                            await atoms[spawned_event.target].put(spawned_event)

            for task in tasks:
                task.cancel()

            await asyncio.gather(*tasks, return_exceptions=True)

        except asyncio.CancelledError:
            for task in tasks:
                task.cancel()

            try:
                await asyncio.wait_for(
                    asyncio.gather(*tasks, return_exceptions=True), timeout=4
                )
            except asyncio.TimeoutError:
                pass

            await self.transport.change_assignation(
                assignation.assignation, status=AssignationStatus.CANCELLED
            )

        except Exception as e:
            logging.critical(f"Assignation {assignation} failed", exc_info=True)

            await self.aass_log(
                assignation.assignation, message=repr(e), level=AssignationStatus.ERROR
            )
            await self.transport.change_assignation(
                assignation.assignation,
                status=AssignationStatus.CRITICAL,
                message=repr(e),
            )

    async def on_unprovide(self):
        for contract in self.contracts.values():
            await contract.aexit()



from typing import Protocol, runtime_checkable
from rekuest.postmans.utils import RPCContract, arkiuse, localuse, mockuse
from fluss.api.schema import (
    ArkitektNodeFragment,
    FlowNodeFragmentBaseArkitektNode,
    LocalNodeFragment,
)
from rekuest.api.schema import afind, ReserveBindsInput
from rekuest.postmans.vars import get_current_postman
from rekuest.structures.registry import get_current_structure_registry
from rekuest.actors.base import Actor


@runtime_checkable
class NodeContractor(Protocol):
    async def __call__(self, node: ArkitektNodeFragment, actor: Actor) -> RPCContract:
        ...


async def arkicontractor(node: ArkitektNodeFragment, actor: Actor) -> RPCContract:
    arkitekt_node = await afind(hash=node.hash)

    return arkiuse(
        binds=ReserveBindsInput(
            clients=node.binds.clients, templates=node.binds.templates
        )
        if node.binds
        else None,
        definition=arkitekt_node,
        postman=get_current_postman(),
        structure_registry=get_current_structure_registry(),
        provision=actor.provision.guardian,
        shrink_inputs=False,
        expand_outputs=False,
        reference=node.id,
        state_hook=actor.on_reservation_change,
    )  # No need to shrink inputs/outsputs for arkicontractors


async def localcontractor(node: LocalNodeFragment, actor: Actor) -> RPCContract:
    print("Creating local contractor node hash: ", node)
    return localuse(
        hash=node.hash,
        agent=actor.agent,
        structure_registry=get_current_structure_registry(),
        shrink_inputs=False,
        expand_outputs=False,
        reference=node.id,
        provision=actor.provision.provision,
    )  # No need to shrink inputs/outputs for arkicontractors


async def arkimockcontractor(node: ArkitektNodeFragment, actor: Actor) -> RPCContract:
    return mockuse(
        node=node,
        provision=actor.provision.guardian,
        shrink_inputs=False,
        shrink_outputs=False,
    )  # No need to shrink inputs/outputs for arkicontractors



import asyncio

print(asyncio.Queue)
from typing import Any, List, Optional
from rekuest.postmans.utils import RPCContract

from fluss.api.schema import ArkitektNodeFragment

from reaktion.atoms.generic import MapAtom, MergeMapAtom, AsCompletedAtom, OrderedAtom
from reaktion.events import Returns
import logging

logger = logging.getLogger(__name__)


class ArkitektMapAtom(MapAtom):
    node: ArkitektNodeFragment
    contract: RPCContract

    async def map(self, args: Returns) -> Optional[List[Any]]:
        defaults = self.node.defaults or {}

        stream_one = self.node.instream[0]
        for arg, item in zip(args, stream_one):
            defaults[item.key] = arg

        returns = await self.contract.aassign(
            **defaults, parent=self.assignation.assignation
        )
        return returns
        # return await self.contract.aassign(*args)


class ArkitektMergeMapAtom(MergeMapAtom):
    node: ArkitektNodeFragment
    contract: RPCContract

    async def merge_map(self, args: Returns) -> Optional[List[Any]]:
        defaults = self.node.defaults or {}

        stream_one = self.node.instream[0]
        for arg, item in zip(args, stream_one):
            defaults[item.key] = arg

        async for r in self.contract.astream(
            **defaults, parent=self.assignation.assignation
        ):
            yield r


class ArkitektAsCompletedAtom(AsCompletedAtom):
    node: ArkitektNodeFragment
    contract: RPCContract

    async def map(self, args: Returns) -> Optional[List[Any]]:
        defaults = self.node.defaults or {}

        stream_one = self.node.instream[0]
        for arg, item in zip(args, stream_one):
            defaults[item.key] = arg

        returns = await self.contract.aassign(
            **defaults, parent=self.assignation.assignation
        )
        return returns


class ArkitektOrderedAtom(OrderedAtom):
    node: ArkitektNodeFragment
    contract: RPCContract

    async def map(self, args: Returns) -> Optional[List[Any]]:
        defaults = self.node.defaults or {}

        stream_one = self.node.instream[0]
        for arg, item in zip(args, stream_one):
            defaults[item.key] = arg

        returns = await self.contract.aassign(
            **defaults, parent=self.assignation.assignation
        )
        return returns



from typing import List, Tuple, Union, List, Tuple, Any
from pydantic import BaseModel, Field, validator
from enum import Enum


class EventType(str, Enum):
    NEXT = "NEXT"
    ERROR = "ERROR"
    COMPLETE = "COMPLETE"


Returns = Tuple[Any, ...]


class InEvent(BaseModel):
    target: str
    """The node that is targeted by the event"""
    handle: str = Field(..., description="The handle of the port")
    """ The handle of the port that emitted the event"""
    type: EventType = Field(..., description="The event type")
    """ The type of event"""
    value: Union[Exception, Returns] = Field(
        None, description="The value of the event (null, exception or any"
    )
    """ The attached value of the event"""
    current_t: int
    """ The current (in loop) time of the event"""

    @validator("handle")
    def validate_handle(cls, v):
        if isinstance(v, int):
            v = f"arg_{v}"

        if v.startswith("return_"):
            raise ValueError(f"Handle needs to start with arg_. This is an inevent {v}")
        if not v.startswith("arg_"):
            raise ValueError(
                f"Handle needs to start with arg_. This is an outevent {v} "
            )

        return v

    class Config:
        arbitrary_types_allowed = True


class OutEvent(BaseModel):
    source: str
    """ The node that emitted the event """
    handle: str = Field(..., description="The handle of the port")
    """ The handle of the port that emitted the event"""
    type: EventType = Field(..., description="The event type")
    """ The type of event"""
    value: Union[Exception, Returns] = Field(
        None, description="The value of the event (null, exception or any"
    )
    caused_by: Tuple[int, ...]
    """ The attached value of the event"""

    @validator("handle")
    def validate_handle(cls, v):
        if isinstance(v, int):
            v = f"return_{v}"

        if v.startswith("arg_"):
            raise ValueError(f"Handle cannot start with arg_. This is an outevent {v}")
        if not v.startswith("return_"):
            raise ValueError(
                f"Handle needs to start with return_. This is an outevent {v}"
            )

        return v

    @validator("value", pre=True)
    def validate_value(cls, v, **kwargs):
        if isinstance(v, Exception):
            return v

        return tuple(v)

    @validator("caused_by", pre=True)
    def validate_caused_by(cls, v, **kwargs):
        return tuple(v)

    def to_state(self):
        if self.value:
            value = (
                self.value if not isinstance(self.value, Exception) else str(self.value)
            )
        else:
            value = None

        return {
            "source": self.source,
            "handle": self.handle,
            "type": self.type,
            "value": value,
        }

    class Config:
        arbitrary_types_allowed = True



import asyncio
from typing import List, Tuple, Optional
from rekuest.api.schema import AssignationLogLevel
from reaktion.atoms.combination.base import CombinationAtom
from reaktion.events import EventType, OutEvent, InEvent
import logging
from pydantic import Field
import functools
from reaktion.atoms.helpers import index_for_handle

logger = logging.getLogger(__name__)


class CombineLatestAtom(CombinationAtom):
    state: List[Optional[InEvent]] = Field(default_factory=lambda: [None, None])

    async def run(self):
        try:
            while True:
                event = await self.get()

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break

                streamIndex = index_for_handle(event.handle)
                print(streamIndex)

                if event.type == EventType.COMPLETE:
                    if streamIndex == 0:
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.COMPLETE,
                                source=self.node.id,
                                caused_by=[event.current_t],
                            )
                        )
                        break

                if event.type == EventType.NEXT:
                    self.state[streamIndex] = event

                    if all(map(lambda x: x is not None, self.state)):
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.NEXT,
                                value=functools.reduce(
                                    lambda a, b: a.value + b.value, self.state
                                ),
                                source=self.node.id,
                                caused_by=map(lambda x: x.current_t, self.state),
                            )
                        )

        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e

        except Exception:
            logger.exception(f"Atom {self.node} excepted")



import asyncio
from typing import List
from reaktion.atoms.combination.base import CombinationAtom
from reaktion.events import EventType, OutEvent
import logging

logger = logging.getLogger(__name__)


class ChunkAtom(CombinationAtom):
    complete: List[bool] = [False, False]

    async def run(self):
        try:
            while True:
                event = await self.get()

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break

                if event.type == EventType.NEXT:
                    assert (
                        len(event.value) == 1
                    ), "ChunkAtom only supports flattening one value"
                    assert isinstance(
                        event.value[0], list
                    ), "ChunkAtom only supports flattening lists"

                    if self.node.defaults:
                        iterations = self.node.defaults.get("iterations", 1)
                    else:
                        iterations = 1

                    for i in range(iterations):
                        for value in event.value[0]:
                            await self.transport.put(
                                OutEvent(
                                    handle="return_0",
                                    type=EventType.NEXT,
                                    value=[value],
                                    source=self.node.id,
                                    caused_by=[event.current_t],
                                )
                            )

                            if self.node.defaults:
                                sleep = self.node.defaults.get("sleep", None)
                                print("Sleeping in interval", sleep)
                                if sleep:
                                    await asyncio.sleep(sleep * 0.001)

                        if self.node.defaults:
                            sleep = self.node.defaults.get("iterations_sleep", None)
                            if sleep:
                                await asyncio.sleep(sleep * 0.001)

                if event.type == EventType.COMPLETE:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            value=[],
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break

        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e

        except Exception as e:
            logger.exception(f"Atom {self.node} excepted")
            raise e



import asyncio
from reaktion.events import OutEvent, Returns, EventType
from reaktion.atoms.base import Atom
import logging
from pydantic import Field
from typing import Dict, List
import asyncio

logger = logging.getLogger(__name__)


class MapAtom(Atom):
    async def map(self, value: Returns) -> Returns:
        raise NotImplementedError("This needs to be implemented")

    async def run(self):
        try:
            while True:
                event = await self.get()

                if event.type == EventType.NEXT:
                    try:
                        result = await self.map(event.value)
                        if result is None:
                            value = ()
                        elif isinstance(result, list) or isinstance(result, tuple):
                            value = result
                        else:
                            value = (result,)

                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.NEXT,
                                value=value,
                                source=self.node.id,
                                caused_by=[event.current_t],
                            )
                        )
                    except Exception as e:
                        logger.error(f"{self.node.id} map failed")
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.ERROR,
                                source=self.node.id,
                                value=e,
                                caused_by=[event.current_t],
                            )
                        )
                        break

                if event.type == EventType.COMPLETE:
                    # Everything left of us is done, so we can shut down as well
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break  # Everything left of us is done, so we can shut down as well

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break
                    # We are not raising the exception here but monadicly killing it to the
                    # left
        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e


class MergeMapAtom(Atom):
    async def merge_map(self, value: Returns) -> Returns:
        raise NotImplementedError("This needs to be implemented")

    async def run(self):
        try:
            while True:
                event = await self.get()

                if event.type == EventType.NEXT:
                    try:
                        async for result in self.merge_map(event.value):
                            if result is None:
                                value = ()
                            elif isinstance(result, list) or isinstance(result, tuple):
                                value = result
                            else:
                                value = (result,)
                            await self.transport.put(
                                OutEvent(
                                    handle="return_0",
                                    type=EventType.NEXT,
                                    value=value,
                                    source=self.node.id,
                                    caused_by=[event.current_t],
                                )
                            )
                    except Exception as e:
                        logger.error(f"{self.node.id} map failed")
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.ERROR,
                                source=self.node.id,
                                value=e,
                                caused_by=[event.current_t],
                            )
                        )
                        break

                if event.type == EventType.COMPLETE:
                    # Everything left of us is done, so we can shut down as well
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break  # Everything left of us is done, so we can shut down as well

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break
                    # We are not raising the exception here but monadicly killing it to the
                    # left
        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e


class OrderedAtom(Atom):
    runningEvents: Dict[int, asyncio.Task] = Field(default_factory=dict)
    publish_queue: List[int] = Field(default_factory=list)

    async def map(self, value: Returns) -> Returns:
        raise NotImplementedError("This needs to be implemented")

    async def publish_changes(self):
        tasks_to_remove = []
        while True:
            await asyncio.sleep(0.1)
            for key, task in self.runningEvents.items():
                if task.done():
                    exception = task.exception()
                    if exception:
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.ERROR,
                                value=exception,
                                source=self.node.id,
                                caused_by=[key],
                            )
                        )
                    else:
                        if key in self.publish_queue:
                            if self.publish_queue[0] == key:
                                await self.transport.put(
                                    OutEvent(
                                        handle="return_0",
                                        type=EventType.NEXT,
                                        value=task.result(),
                                        source=self.node.id,
                                        caused_by=[key],
                                    )
                                )
                                self.publish_queue.pop(0)
                                tasks_to_remove.append(key)

            for key in tasks_to_remove:
                self.runningEvents.pop(key)

            tasks_to_remove = []

    async def run(self):
        publish_task = asyncio.create_task(self.publish_changes())

        try:
            while True:
                event = await self.get()

                if event.type == EventType.NEXT:
                    self.publish_queue.append(event.current_t)
                    self.runningEvents[event.current_t] = asyncio.create_task(
                        self.map(event.value)
                    )

                if event.type == EventType.COMPLETE:
                    # Everything left of us is done, so we can shut down as well
                    await asyncio.gather(*self.runningEvents.values())
                    publish_task.cancel()
                    try:
                        await publish_task
                    except asyncio.CancelledError:
                        pass
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break  # Everything left of us is done, so we can shut down as well

                if event.type == EventType.ERROR:
                    for key, value in self.runningEvents.items():
                        value.cancel()
                        try:
                            await value
                        except asyncio.CancelledError:
                            pass

                    publish_task.cancel()
                    try:
                        await publish_task
                    except asyncio.CancelledError:
                        pass

                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break
                    # We are not raising the exception here but monadicly killing it to the
                    # left
        except asyncio.CancelledError as e:
            publish_task.cancel()
            try:
                await publish_task
            except asyncio.CancelledError:
                pass

            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e


class AsCompletedAtom(Atom):
    runningEvents: Dict[int, asyncio.Task] = Field(default_factory=dict)

    async def map(self, value: Returns) -> Returns:
        raise NotImplementedError("This needs to be implemented")

    async def publish_changes(self):
        tasks_to_remove = []
        while True:
            await asyncio.sleep(0.1)
            for key, task in self.runningEvents.items():
                if task.done():
                    exception = task.exception()
                    if exception:
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.ERROR,
                                value=exception,
                                source=self.node.id,
                                caused_by=[key],
                            )
                        )

                    else:
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.NEXT,
                                value=task.result(),
                                source=self.node.id,
                                caused_by=[key],
                            )
                        )
                    tasks_to_remove.append(key)

            for key in tasks_to_remove:
                self.runningEvents.pop(key)

            tasks_to_remove = []

    async def run(self):
        publish_task = asyncio.create_task(self.publish_changes())

        try:
            while True:
                event = await self.get()

                if event.type == EventType.NEXT:
                    self.runningEvents[event.current_t] = asyncio.create_task(
                        self.map(event.value)
                    )

                if event.type == EventType.COMPLETE:
                    # Everything left of us is done, so we can shut down as well
                    await asyncio.gather(*self.runningEvents.values())
                    publish_task.cancel()
                    try:
                        await publish_task
                    except asyncio.CancelledError:
                        pass
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break  # Everything left of us is done, so we can shut down as well

                if event.type == EventType.ERROR:
                    for key, value in self.runningEvents.items():
                        value.cancel()
                        try:
                            await value
                        except asyncio.CancelledError:
                            pass

                    publish_task.cancel()
                    try:
                        await publish_task
                    except asyncio.CancelledError:
                        pass

                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            value=event.value,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break
                    # We are not raising the exception here but monadicly killing it to the
                    # left
        except asyncio.CancelledError as e:
            publish_task.cancel()
            try:
                await publish_task
            except asyncio.CancelledError:
                pass

            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e



from typing import Awaitable, Callable, Dict
from rekuest.api.schema import AssignationLogLevel, NodeKind
from rekuest.messages import Assignation
from fluss.api.schema import (
    ArkitektNodeFragment,
    LocalNodeFragment,
    FlowNodeFragment,
    ReactiveImplementationModelInput,
    ReactiveNodeFragment,
    MapStrategy,
)
import asyncio

print(asyncio.Queue)

from reaktion.atoms.arkitekt import (
    ArkitektMapAtom,
    ArkitektMergeMapAtom,
    ArkitektAsCompletedAtom,
    ArkitektOrderedAtom,
)
from reaktion.atoms.local import LocalMapAtom, LocalMergeMapAtom
from reaktion.atoms.transformation.chunk import ChunkAtom
from reaktion.atoms.transformation.split import SplitAtom
from reaktion.atoms.combination.zip import ZipAtom
from reaktion.atoms.combination.withlatest import WithLatestAtom
from rekuest.postmans.utils import RPCContract
from .base import Atom
from .transport import AtomTransport


def atomify(
    node: FlowNodeFragment,
    transport: AtomTransport,
    contracts: Dict[str, RPCContract],
    local_contracts: Dict[str, RPCContract],
    assignation: Assignation,
    alog: Callable[[Assignation, AssignationLogLevel, str], Awaitable[None]] = None,
) -> Atom:
    if isinstance(node, ArkitektNodeFragment):
        if node.kind == NodeKind.FUNCTION:
            if node.map_strategy == MapStrategy.MAP:
                return ArkitektMapAtom(
                    node=node,
                    contract=contracts[node.id],
                    transport=transport,
                    assignation=assignation,
                    alog=alog,
                )
            if node.map_strategy == MapStrategy.AS_COMPLETED:
                return ArkitektAsCompletedAtom(
                    node=node,
                    contract=contracts[node.id],
                    transport=transport,
                    assignation=assignation,
                    alog=alog,
                )
            if node.map_strategy == MapStrategy.ORDERED:
                return ArkitektAsCompletedAtom(
                    node=node,
                    contract=contracts[node.id],
                    transport=transport,
                    assignation=assignation,
                    alog=alog,
                )
        if node.kind == NodeKind.GENERATOR:
            return ArkitektMergeMapAtom(
                node=node,
                contract=contracts[node.id],
                transport=transport,
                assignation=assignation,
                alog=alog,
            )

    if isinstance(node, LocalNodeFragment):
        if node.kind == NodeKind.FUNCTION:
            return LocalMapAtom(
                node=node,
                contract=local_contracts[node.id],
                transport=transport,
                assignation=assignation,
                alog=alog,
            )
        if node.kind == NodeKind.GENERATOR:
            return LocalMergeMapAtom(
                node=node,
                contract=local_contracts[node.id],
                transport=transport,
                assignation=assignation,
                alog=alog,
            )

    if isinstance(node, ReactiveNodeFragment):
        if node.implementation == ReactiveImplementationModelInput.ZIP:
            return ZipAtom(
                node=node,
                transport=transport,
                assignation=assignation,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.CHUNK:
            return ChunkAtom(
                node=node,
                transport=transport,
                assignation=assignation,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.WITHLATEST:
            return WithLatestAtom(
                node=node,
                transport=transport,
                assignation=assignation,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.COMBINELATEST:
            return WithLatestAtom(
                node=node,
                transport=transport,
                assignation=assignation,
                alog=alog,
            )
        if node.implementation == ReactiveImplementationModelInput.SPLIT:
            return SplitAtom(
                node=node,
                transport=transport,
                assignation=assignation,
                alog=alog,
            )

    raise NotImplementedError(f"Atom for {node} is not implemented")
